# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Authentication Configuration Typing """
import enum
from dataclasses import dataclass
from typing import Literal


class UserRole(str, enum.Enum):
    """User Roles

    Args:
        enum (str): Available types of user roles:
        * user
        * admin
    """

    USER = "user"
    ADMIN = "admin"


class GrantType(str, enum.Enum):
    """Grant Type

    Args:
        enum (str): the only available grant type is urn:ietf:params:oauth:grant-type:jwt-bearer
    """

    JWT_BEARER = "urn:ietf:params:oauth:grant-type:jwt-bearer"


class UserScope(str, enum.Enum):
    """User Scope

    Args:
        enum (str): the only available scope is 'user profile'
    """

    USER = "user profile"


@dataclass
class AuthorisationRequestPayload:
    """Authorisation Request Payload

    Attributes:
        grantType (str): the grant type
        assertion (str): assertion
        scope (UserScope): User scope

    """

    grantType: Literal[GrantType.JWT_BEARER]  # pylint: disable=invalid-name
    assertion: str
    scope: UserScope


class TokenType(str, enum.Enum):
    """Token Type

    Args:
        enum (str): only available token type is 'bearer'
    """

    BEARER = "bearer"


@dataclass
class AccessTokenPartialPayload:
    """Access Token Partial Payload

    Attributes:
        user_id (int): the user id
        user_role (UserRole): user role
        aud: target audience of the token
    """

    user_id: int
    user_role: UserRole
    aud: str


@dataclass
class AccessTokenResponse:
    """Access Token Response

    Attributes:
        accessToken (str): the access token
        tokenType (str): the token type
        expiresIn (int): timestamp when the token expires
        issuedAt (str): string representation of the timestamp when token was issued
    """

    accessToken: str  # pylint: disable=invalid-name
    refreshToken: str | None  # pylint: disable=invalid-name
    tokenType: Literal[TokenType.BEARER]  # pylint: disable=invalid-name
    expiresIn: int  # pylint: disable=invalid-name
    issuedAt: str  # pylint: disable=invalid-name


@dataclass
class AssertionPayload:
    """Assertion Payload

    Attributes:
        user_id (int): the user id
        username (str): the user name
        api_key (str): the api key associated to the user
        audience (str): target audience where token is used
        iat (int): timestamp when token was issued
    """

    user_id: int
    username: str
    api_key: str
    audience: str
    iat: int



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Device class """
import json

from typeguard import typechecked

from qiboconnection.connection import Connection
from qiboconnection.typings.devices import DeviceInput
from qiboconnection.typings.enums import DeviceStatus

from .device_details import DeviceDetails

# pylint: disable=no-member
# pylint: disable=attribute-defined-outside-init
# pylint: disable=too-many-instance-attributes


class Device(DeviceDetails):
    """Abstract class for devices."""

    @typechecked
    def __init__(self, device_input: DeviceInput):
        super().__init__()
        for k, v in vars(device_input).items():
            setattr(self, f"_{k}", v)
        self._str = (
            f"<Device: id={self._id}, name='{self._name}', "  # type: ignore[attr-defined]
            + f"status='{self._status}', type='{self._type}'>"  # type: ignore[attr-defined]
        )

    def __str__(self):
        """String representation of a Device

        Returns:
            str: String representation of a Device
        """
        return self._str

    def __repr__(self) -> str:
        return self.__str__()

    @property
    def id(self) -> int:  # pylint: disable=invalid-name
        """Returns device identifier

        Returns:
            int: device identifier
        """
        return self._id  # type: ignore[attr-defined]

    @property
    def name(self) -> str:
        """Returns device name

        Returns:
            str: device name
        """
        return self._name  # type: ignore[attr-defined]

    @property
    def status(self) -> DeviceStatus | None:
        """Returns device name

        Returns:
            str: device name
        """
        _status = getattr(self, "_status", None)
        return _status  # type: ignore[attr-defined]

    @property
    def type(self) -> str:
        """Returns device type

        Returns:
            str: device type
        """
        return getattr(self, "_type")

    @property
    def number_pending_jobs(self) -> int | None:
        """Return pending jobs in the queue for the requested device

        Returns:
            int: pending jobs

        """
        return getattr(self, "_number_pending_jobs", None)

    @property
    def static_features(self) -> int | None:
        """Return pending jobs in the queue for the requested device

        Returns:
            int: pending jobs

        """
        return getattr(self, "_static_features", None)

    @property
    def dynamic_features(self) -> int | None:
        """Return pending jobs in the queue for the requested device

        Returns:
            int: pending jobs

        """
        return getattr(self, "_dynamic_features", None)

    def set_to_online(self, connection: Connection) -> None:
        """Updates a device status so that it can accept remote jobs. Local jobs will be blocked.

        Args:
            connection (Connection): Qibo API connection
        """
        connection.update_device_status(device_id=self._id, status=DeviceStatus.ONLINE)  # type: ignore[attr-defined]
        self._status = DeviceStatus.ONLINE

    def set_to_maintenance(self, connection: Connection) -> None:
        """Puts a device in maintenance mode, so that it can only accept local jobs.
         Incoming remote jobs will be blocked.

        Args:
            connection (Connection): Qibo API connection
        """
        connection.update_device_status(device_id=self._id, status=DeviceStatus.MAINTENANCE)  # type: ignore[attr-defined]
        self._status = DeviceStatus.MAINTENANCE

    def to_dict(self, expand=False):
        """Dictionary representation of a Device

        Returns:
            dict: Output dictionary of a Device object
        """
        return self._get_expanded_dict() if expand else self._get_truncated_dict()

    def _get_truncated_dict(self):
        """
        Builds a dict representation for the device that contains only the most common checked fields.
        Fields that are not json serializable are ignored.
        Returns: dict
        """
        result_dict = {}
        truncated_fields = {"id", "name", "status", "type", "number_pending_jobs"}
        for key in truncated_fields:
            value = getattr(self, f"_{key}", None)
            try:
                json.dumps(value)
                result_dict |= {key: getattr(self, key, None)}
            except (json.JSONDecodeError, TypeError):
                pass
        return result_dict

    def _get_expanded_dict(self):
        """
        Builds a dict representation for the device that contains all the fields present in the device, including those
        that qiboconnection does not know about.
        Fields that are not json serializable are ignored.
        Returns: dict
        """
        result_dict = {}
        for key in dir(self):
            if len(key) >= 2 and key[0] == "_" and key[1] != "_":
                value = getattr(self, key, None)
                try:
                    json.dumps(value)
                    result_dict |= {key[1:]: getattr(self, key, None)}
                except (json.JSONDecodeError, TypeError):
                    pass
        return result_dict

    def toJSON(self, expand=False) -> str:  # pylint: disable=invalid-name
        """JSON representation of a Device

        Returns:
            str: JSON serialization of a Device object
        """

        return json.dumps(self.to_dict(expand=expand), indent=2)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" JobResult typing """

import json
import logging
from abc import ABC
from dataclasses import dataclass, field
from typing import List

from numpy import typing as npt
from qibo.result import CircuitResult

from qiboconnection.typings.enums import JobType
from qiboconnection.util import decode_results_from_circuit, decode_results_from_qprogram, decompress_any

logger = logging.getLogger(__name__)


@dataclass
class JobResult(ABC):
    """Job Result class"""

    job_id: int
    http_response: str
    job_type: str
    data: List[CircuitResult] | CircuitResult | npt.NDArray | List[int] | List[float] | dict | List[
        dict
    ] | str | None = field(init=False)

    def __post_init__(self) -> None:
        """
        Decodes data from the http_response provided at instance creation, and uses that info to build the self.data
        attribute.
        """
        try:
            self.data = decompress_any(**json.loads(self.http_response))
            return
        except Exception as ex:  # pylint: disable=broad-exception-caught  # delete ASAP
            logger.warning(
                f"Unable to decompress JobResult with get_results interface due to {ex}({type(ex)})."
                f"Falling back to legacy methods."
            )

            if self.job_type == JobType.CIRCUIT:
                self.data = decode_results_from_circuit(self.http_response)
                return
            if self.job_type == JobType.QPROGRAM:
                self.data = decode_results_from_qprogram(self.http_response)
                return
            if self.job_type == JobType.ANNEALING_PROGRAM:
                self.data = decode_results_from_qprogram(self.http_response)
                return
            if self.job_type == JobType.VQA:
                self.data = decode_results_from_qprogram(self.http_response)
                return

            logger.warning("Result not supported for type of job. Returning a plain string. ")
            self.data = self.http_response



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CONSTANTS"""


class REST:  # pylint: disable=too-few-public-methods
    """REST constants, for rest calls used in many sites"""

    ITEMS = "items"


class REST_ERROR:  # pylint: disable=too-few-public-methods
    """REST error constants"""

    DETAIL = "detail"


class API_CONSTANTS:  # pylint: disable=too-few-public-methods
    """API constants"""

    JOB_ID = "job_id"
    USER_ID = "user_id"
    SAVED_EXPERIMENT_ID = "saved_experiment_id"
    RUNCARD_ID = "runcard_id"
    FAVOURITE = "favourite"
    FAVOURITES = "favourites"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Job Typing """
import json
from abc import ABC
from dataclasses import asdict, dataclass, field
from typing import Any, List

from qibo.models.circuit import Circuit
from typeguard import typechecked

from qiboconnection.typings.enums import JobStatus, JobType
from qiboconnection.typings.requests import JobRequest
from qiboconnection.typings.responses.job_response import JobResponse
from qiboconnection.typings.vqa import VQA
from qiboconnection.util import compress_any

from .algorithm import ProgramDefinition
from .devices.device import Device
from .job_result import JobResult
from .user import User


@dataclass
class Job(ABC):  # pylint: disable=too-many-instance-attributes
    """Job class to manage the job experiment to be remotely sent"""

    user: User
    device: Device
    program: ProgramDefinition | None = field(default=None)
    circuit: list[Circuit] | None = None
    qprogram: str | None = None
    anneal_program_args: dict | None = None
    vqa: VQA | None = None
    nshots: int = 10
    job_status: JobStatus = JobStatus.NOT_SENT
    job_result: JobResult | None = None
    name: str = "-"
    summary: str = "-"
    id: int = 0  # pylint: disable=invalid-name

    def __post_init__(self):
        n = len([arg for arg in [self.qprogram, self.circuit, self.anneal_program_args, self.vqa] if arg is not None])
        match n:
            case n if n > 1:
                raise ValueError(
                    "VQA, circuit, qprogram and anneal_program_args were provided, but execute() only takes one of them."
                )
            case 0:
                raise ValueError("Neither of circuit, vqa, qprogram or anneal_program_args were provided.")

    @property
    def user_id(self) -> int | None:
        """User identifier

        Returns:
            int: User identifier
        """
        return self.user.user_id

    @property
    def device_id(self) -> int:
        """Device identifier

        Returns:
            int: Device identifier
        """
        return self.device.id

    @property
    def algorithms(self) -> List[dict]:
        """Get all algorithms described


        Returns:
            List[dict]: a list of all algorithms as a dictionary
        """
        if self.program is None:
            raise ValueError("Job does not contain an algorithm Program")

        return [algorithm.__dict__ for algorithm in self.program.algorithms]

    @property
    def job_request(self) -> JobRequest:
        """Returns a Job Request with the Job instance info

        Returns:
            JobRequest: Job Request object
        """
        return JobRequest(
            user_id=self.user.user_id,
            device_id=self.device.id,
            number_shots=self.nshots,
            job_type=self.job_type,
            name=self.name,
            summary=self.summary,
            description=self._get_job_description(),
        )

    @property
    def job_id(self) -> int:
        """Returns Job identifier

        Returns:
            int: Job identifier
        """
        return self.id

    @job_id.setter
    def job_id(self, job_id: int) -> None:
        """
        Modifies internal Job identifier, accessed via job_id property
        Args:
            job_id: new job_id value
        """
        self.id = job_id

    @property
    def job_type(self):
        """Get the type of the job, checking whether the user has defined circuit or experiment."""
        if self.circuit is not None and self.qprogram is None and self.anneal_program_args is None and self.vqa is None:
            return JobType.CIRCUIT
        if self.qprogram is not None and self.circuit is None and self.anneal_program_args is None and self.vqa is None:
            return JobType.QPROGRAM
        if self.anneal_program_args is not None and self.circuit is None and self.qprogram is None and self.vqa is None:
            return JobType.ANNEALING_PROGRAM
        if self.vqa is not None and self.circuit is None and self.qprogram is None and self.anneal_program_args is None:
            return JobType.VQA
        raise ValueError("Could not determine JobType")

    def _get_job_description(self) -> str:
        """Serialize either circuit or qprogram to obtain job description"""

        if self.qprogram is not None:
            return json.dumps(compress_any(self.qprogram))
        if self.anneal_program_args is not None:
            return json.dumps(compress_any(self.anneal_program_args))
        if self.vqa is not None:
            vqa_as_dict = asdict(self.vqa)
            vqa_as_dict.pop("vqa_dict")
            return json.dumps({**compress_any(self.vqa.vqa_dict), **vqa_as_dict})
        if self.circuit is not None:
            return json.dumps(compress_any([c.to_qasm() for c in self.circuit]))

        raise ValueError("No suitable information found for building description.")

    @property
    def result(self) -> Any:
        """Return the result of the job

        Raises:
            ValueError: Job result still not completed

        Returns:
            Any: Job Result data
        """
        if self.job_result is None:
            raise ValueError("Job result still not completed")
        return self.job_result.data

    @typechecked
    def update_with_job_response(self, job_response: JobResponse) -> None:
        """Updates the current job with the given job response

        Args:
            job_response (JobResponse): Job response

        Raises:
            ValueError: Job response does not belong to the user.
            ValueError: Job response does not belong to the device.
        """
        if self.user.user_id != job_response.user_id:
            raise ValueError("Job response does not belong to the user.")
        if self.device.id != job_response.device_id:
            raise ValueError("Job response does not belong to the device.")
        self.job_id = job_response.job_id
        self.job_status = (
            job_response.status if isinstance(job_response.status, JobStatus) else JobStatus(job_response.status)
        )
        self.job_result = JobResult(job_id=self.id, job_type=job_response.job_type, http_response=job_response.result)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Runcard typing classes """
from dataclasses import dataclass
from datetime import datetime

from qiboconnection.util import from_kwargs

from ..requests import RuncardRequest


@dataclass
class RuncardResponse(RuncardRequest):
    """Class for accommodating Runcards."""

    runcard_id: int
    created_at: datetime
    updated_at: datetime

    @classmethod
    def from_kwargs(cls, **kwargs):
        "Returns an instance of RuncardResponse including non-typed attributes"
        return from_kwargs(cls, **kwargs)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Connection Typing """

from abc import ABC
from dataclasses import dataclass


@dataclass
class _ConnectionConfigurationDefaultBase(ABC):
    """Internal class for the default-value attributes of ConnectionConfiguration"""

    user_id: int | None = None


@dataclass
class _ConnectionConfigurationBase(ABC):
    """Internal class for the non-default-value attributes of ConnectionConfiguration"""

    username: str
    api_key: str


@dataclass
class ConnectionConfiguration(_ConnectionConfigurationDefaultBase, _ConnectionConfigurationBase):
    """Connection Configuration

    Attributes:
        user_id (int): The user id
        username (str): The user name
        api_key (str): The API key associated to the user
    """


@dataclass
class _ConnectionEstablishedBase(ABC):
    """Internal class for the non-default-value attributes of ConnectionEstablished"""

    authorisation_access_token: str
    authorisation_refresh_token: str
    api_path: str


@dataclass
class ConnectionEstablished(ConnectionConfiguration, _ConnectionEstablishedBase):
    """Connection Established

    Attributes:
        user_id (int): The user id
        username (str): The user name
        api_key (str): The API key associated to the user
        authorisation_access_token (str): access token associated
        authorisation_refresh_token (str): refresh token associated
        api_path (str): API path to use the token
    """



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" JobResponse """
from dataclasses import dataclass

from qiboconnection.util import from_kwargs

from ..requests import JobRequest


@dataclass
class JobResponse(JobRequest):
    """Full Job Response. Includes job results which may
    be weight a few GB.

    Attributes:
        user_id (int): User identifier
        device_id (int): Device identifier
        description (str): Description of the job
        job_id (int): Job identifier
        queue_position (int): Job queue position
        status (str): Status of the job
        result (str): Job result
    """

    job_id: int
    queue_position: int
    result: str
    status: str
    name: str
    summary: str

    @classmethod
    def from_kwargs(cls, **kwargs):
        """Returns an instance of JobResponse including non-typed attributes"""
        return from_kwargs(cls, **kwargs)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

r"""
This module contains a function to display all the details of the Qiboconnection installation.
"""
import platform
import sys
from subprocess import check_output  # nosec B404

import qibo


def about():
    """
    Prints the information for qiboconnection installation.
    """
    print(check_output([sys.executable, "-m", "pip", "show", "qiboconnection"]).decode())  # nosec B603
    print(f"Platform info:           {platform.platform(aliased=True)}")
    print(f"Python version:          {sys.version_info[0]}.{sys.version_info[1]}.{sys.version_info[2]}")
    print(f"Qibo version:            {qibo.__version__}")



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" DeviceType enum """
from .str_enum import StrEnum


class DeviceType(StrEnum):
    """Device Type

    Args:
        enum (str): Device type options available:
        * quantum
        * simulator
    """

    SIMULATOR_DEVICE = "simulator_device"
    QUANTUM_DEVICE = "quantum_device"
    QUANTUM_ANALOG_DEVICE = "quantum_analog_device"
    QUANTUM_DIGITAL_DEVICE = "quantum_digital_device"



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Device Typing """
from dataclasses import dataclass

from qiboconnection.util import from_kwargs

# pylint: disable=too-many-instance-attributes


@dataclass(kw_only=True)
class DeviceInput:
    """Device Input

    Attributes:
        id (int): device identifier
        name (str): device name
        status (str): device status
    """

    id: int
    name: str
    status: str
    type: str | None = None
    number_pending_jobs: int | None = None
    slurm_partition: str | None = None
    static_features: dict | None = None
    dynamic_features: dict | None = None

    @classmethod
    def from_kwargs(cls, **kwargs):
        "Returns an instance of DeviceInput including non-typed attributes"

        kwargs = cls._apply_retrocompatibility_conversions(kwargs=kwargs)

        return from_kwargs(cls, **kwargs)

    @classmethod
    def _apply_retrocompatibility_conversions(cls, kwargs):
        """
        If old required keys are provided instead of new ones, update the dictionary with the new
        expected values.
        """
        if "device_id" in kwargs:
            kwargs["id"] = kwargs.pop("device_id")
        if "device_name" in kwargs:
            kwargs["name"] = kwargs.pop("device_name")
        if "device_type" in kwargs:
            kwargs["type"] = kwargs.pop("device_type")
        return kwargs



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC
from dataclasses import dataclass


@dataclass
class JobRequest(ABC):
    """Job Request

    Attributes:
        user_id (int): User identifier
        device_id (int): Device identifier
        description (str): Description of the job
        job_type (str | JobType): Type of the job
        number_shots (int): number of times the job is to be executed
    """

    user_id: int | None
    device_id: int
    number_shots: int
    job_type: str
    description: str
    name: str
    summary: str



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Util Functions used by the API module """

import json
from typing import Any, List

from qibo.models.circuit import Circuit

from qiboconnection.config import logger
from qiboconnection.models import JobResult
from qiboconnection.typings.enums import JobStatus, JobType
from qiboconnection.typings.responses.job_response import JobResponse
from qiboconnection.util import decompress_any


def parse_job_responses_to_results(job_responses: List[JobResponse]) -> List[dict | Any | None]:
    """Parse a list of job_responses to a list of dict with the content of each job. If the job is not COMPLETED,
    put a None in its place. For this, we build a JobResult instance for each COMPLETED job, and then we keep its
    `.data`.

    Args:
        job_responses: list of JobResponse instances from which we'll

    Returns:

    """
    raw_results = [parse_job_response_to_result(job_response=job_response) for job_response in job_responses]
    return list(raw_results)


def parse_job_response_to_result(job_response: JobResponse):
    """Parse a single job_response to a single dict with the content of a job. If the job is not COMPLETED,
    put a None in its place. For this, we build a JobResult instance for each COMPLETED job, and then we keep its
    `.data`.

    Args:
        job_response: JobResponse instance from which we'll get the results

    Returns:

    """
    return (
        JobResult(job_id=job_response.job_id, job_type=job_response.job_type, http_response=job_response.result).data
        if job_response.status == JobStatus.COMPLETED
        else None
    )


def deserialize_job_description(raw_description: str, job_type: str) -> dict:
    """Convert base64 job description to its corresponding Qibo Circuit or Qililab experiment

    Args:
        raw_description (str):
        job_type (str):

    Raises:
        ValueError: Job type isn't nor Qibo Circuit neither Qililab experiment

    Returns:
        Circuit | dict: _description_
    """

    description_dict = json.loads(raw_description)
    decompressed_data = decompress_any(**description_dict)
    compressed_data = description_dict.pop("data")
    if job_type == JobType.CIRCUIT:
        return {
            **description_dict,
            "data": [Circuit.from_qasm(decom_data) for decom_data in decompressed_data],
        }
    if job_type == JobType.VQA:
        return {**description_dict, "vqa_dict": decompressed_data}
    if job_type in [JobType.QPROGRAM, JobType.ANNEALING_PROGRAM, JobType.OTHER]:
        return {**description_dict, "data": decompressed_data}
    return {**description_dict, "data": compressed_data}


def log_job_status_info(job_response: JobResponse):
    """Logs a message depending on the `JobResponse.status` belonging to a certain job.

    Args:
        job_response:

    Returns:

    """
    if job_response.status in [JobStatus.QUEUED, JobStatus.PENDING]:
        logger.warning(
            "Your job with id %i is still pending. Job queue position: %s",
            job_response.job_id,
            job_response.queue_position,
        )
        return None

    if job_response.status == JobStatus.RUNNING:
        logger.warning("Your job with id %i is still running.", job_response.job_id)
        return None
    if job_response.status == JobStatus.NOT_SENT:
        logger.warning("Your job with id %i has not been sent.", job_response.job_id)
        return None
    if job_response.status == JobStatus.ERROR:
        logger.error("Your job with id %i failed.", job_response.job_id)
        return None
    if job_response.status == JobStatus.COMPLETED:
        logger.warning("Your job with id %i is completed.", job_response.job_id)
        return None

    logger.warning(f"Your job with id %i is {job_response.status}.", job_response.job_id)
    return None



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" StrEnum base class """
from enum import Enum


class StrEnum(str, Enum):
    """Associates a number to each entry. Unlike a normal enum, this is not used for indexing, and adding the '.value'
    to everywhere is still not needed.
    This int can be used for where a string is not suitable."""

    @property
    def to_int(self):
        """Returns the int associated to the str entry in the enum class. Order is important."""
        return list(self.__class__).index(self)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Device Utility Methods
"""

from typeguard import typechecked

from qiboconnection.typings.devices import DeviceInput

from .device import Device


@typechecked
def create_device(device_input: dict) -> Device:
    """Creates a Device from a given device input.

    Args:
        device_input (dict): Device Input structure

    Returns:
        Device: The constructed Device Object
    """

    return Device(device_input=DeviceInput.from_kwargs(**device_input))



from dataclasses import dataclass


@dataclass
class VQA:
    """
    Class to be filled for executing variational quantum algorithms through Qilimanjaro's QaaS.
    """

    vqa_dict: dict
    init_params: list
    optimizer_params: dict | None = None

    def __post_init__(self):
        if self.optimizer_params is None:
            self.optimizer_params = {}
